from os import getcwd, listdir, path, PathLike, chdir, walk, chmod
from requests import Session
from subprocess import run
from shutil import rmtree
from stat import S_IRWXU

##################################################

# Path to root dir of all repos (pre-populated to the linux default)
base_path_to_repos = "/var/atlassian/application-data/bitbucket/shared/data/repos/"

# Bitbucket Cloud credentials
cloud_username = ""
cloud_app_password = "" # https://support.atlassian.com/bitbucket-cloud/docs/app-passwords/

# Desired Bitbucket Cloud workspace name
cloud_workspace = ""

#####
# Please setup an admin account and generate an ssh key,
# placing the private key on the machine that this script will
# be run from so as to pre-handle auth for git pushes to the cloud
#####
# Git details to use for the push in to cloud
git_username = ""
git_user_email = ""

##################################################

session = Session()
session.auth = (cloud_username, cloud_app_password)
cloud_api = "https://api.bitbucket.org"
base_dir = getcwd()


class Repo():
    def __init__(self, project: str, repo: str, path: PathLike, dir: str):
        self.project_name = project
        self.repo_name = repo
        self.path = path
        self.id = dir

    def set_temp_path(self, path: PathLike):
        self.temp_path = path
        return self


def confirm_workspace():
    r = session.get(f"{cloud_api}/2.0/workspaces/{cloud_workspace}")
    if r.status_code == 200:
        return True
    elif r.status_code == 401:
        exit("[FATAL] Unable to access given workspace due to http 401 error. "
             "Please confirm your credentials and try again. Stopping...")
    elif r.status_code == 404:
        exit(f"[FATAL] Unable to locate the given workspace '{cloud_workspace}' within your account. "
             "Please confirm the spelling of the workspace and try again. Stopping...")
    elif r.status_code == 403:
        exit(f"[FATAL] Unable to access the workspace '{cloud_workspace}' as the given credentials "
             "do not provide sufficient permissions. Stopping...")

def configure_git():
    set_username = ['git', 'config', '--global', 'user.name', git_username]
    run(set_username)
    set_user_email = ['git', 'config', '--global', 'user.email', git_user_email]
    run(set_user_email)

def filesystem_repos() -> Repo:
    for dir in listdir(base_path_to_repos):
        with open(path.join(base_path_to_repos, dir, "repository-config")) as f:
            for line in f:
                if line.split()[0] == "project":
                    project_name = line.split()[2]
                if line.split()[0] == "repository":
                    repo_name = line.split()[2]
        repo = Repo(project_name, repo_name, path.join(base_path_to_repos, dir), dir)
        yield repo

def temp_clone(repo: Repo) -> Repo:
    temp_clone = ['git', 'clone', '--no-local', repo.path, f'temp-{repo.repo_name}']
    run(temp_clone)
    repo = repo.set_temp_path(path.join(base_dir, f'temp-{repo.repo_name}'))
    return repo

def create_cloud_project(project: str):
    r = session.get(f"{cloud_api}/2.0/teams/{cloud_workspace}/projects/{project}")
    if r.status_code == 200:
        return True

    payload = {"name": project}
    r = session.post(f"{cloud_api}/2.0/teams/{cloud_workspace}/projects", json=payload)
    if r.status_code == 201:
        return True
    return False

def create_empty_cloud_repo(repo: Repo):
    payload = {'scm': 'git', 'project': {'key': repo.project_name}}
    r = session.post(f"{cloud_api}/2.0/repositories/{cloud_workspace}/{repo.repo_name}",
                     json=payload)
    if r.status_code != 200:
        with open(path.join(base_dir, 'failure.txt'), 'w+') as f:
            f.write(f"Failed to create repo: '{repo.repo_name}' in project: '{repo.project_name}' with id: {repo.id}, skipping...")
    return r.status_code == 200

def push_repo(repo: Repo):
    chdir(repo.temp_path)
    git_add_remote = ['git', 'remote', 'add', 'bitbucket-cloud',
                      f'git@bitbucket.org:{cloud_workspace}/{repo.repo_name}.git']
    run(git_add_remote)
    git_push = ['git', 'push', '-u', 'bitbucket-cloud', '--all']
    run(git_push)
    git_push_tags = ['git', 'push', '-u', 'bitbucket-cloud', '--tags']
    run(git_push_tags)
    chdir(base_dir)

def cleanup_temp(path: PathLike):
    try:
        # ensure script has perms over file path before deleting
        for root, dirs, files in walk(path):
            for d in dirs:
                chmod(path.join(root, d), S_IRWXU)
            for f in files:
                chmod(path.join(root, f), S_IRWXU)
        # recusrively delete temp directory
        rmtree(path)
    except OSError as e:
        print(f"Unable to delete '{path}' automatically. Manual cleanup is "
              f"necessary.\n{e.strerror}")

def migrate():
    for repo in filesystem_repos():
        repo = temp_clone(repo)
        if create_cloud_project(repo.project_name):
            if create_empty_cloud_repo(repo):
                push_repo(repo)
                cleanup_temp(repo.temp_path)


if __name__ == '__main__':
    confirm_workspace()
    configure_git()
    migrate()
